import pkg_resources
import sys
import os
import shlex

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'repoze.sphinx.autointerface',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'pyramid_authsanity'
copyright = '2015, Bert JW Regeer'
author = 'Bert JW Regeer'

version = release = pkg_resources.get_distribution('pyramid_authsanity').version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

modindex_common_prefix = ['pyramid_authsanity.']

# -- Options for HTML output ----------------------------------------------

html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'pyramid_authsanitydoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'pyramid_authsanity.tex', 'pyramid\\_authsanity Documentation',
   'Bert JW Regeer', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'pyramid_authsanity', 'pyramid_authsanity Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'pyramid_authsanity', 'pyramid_authsanity Documentation',
   author, 'pyramid_authsanity', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}



import pytest

from pyramid.authorization import ACLAuthorizationPolicy
import pyramid.testing

from zope.interface.verify import verifyClass

from pyramid_services import find_service_factory

from pyramid_authsanity.interfaces import (
    IAuthSourceService,
    )

class TestAuthServicePolicyIntegration(object):
    @pytest.fixture(autouse=True)
    def pyramid_config(self, request):
        self.config = pyramid.testing.setUp()
        self.config.set_authorization_policy(ACLAuthorizationPolicy())

        def finish():
            del self.config
            pyramid.testing.tearDown()
        request.addfinalizer(finish)

    def _makeOne(self, settings):
        self.config.registry.settings.update(settings)
        self.config.include('pyramid_authsanity')

    def test_include_me(self):
        from pyramid_authsanity.policy import AuthServicePolicy
        self._makeOne({})
        self.config.commit()
        introspector = self.config.registry.introspector
        auth_policy = introspector.get('authentication policy', None)

        assert isinstance(auth_policy['policy'], AuthServicePolicy)

        with pytest.raises(ValueError):
            find_service_factory(self.config, IAuthSourceService)

    def test_include_me_cookie_no_secret(self):
        settings = {'authsanity.source': 'cookie'}

        with pytest.raises(RuntimeError):
            self._makeOne(settings)

    def test_include_me_cookie_with_secret(self):
        from pyramid_authsanity.policy import AuthServicePolicy
        settings = {'authsanity.source': 'cookie', 'authsanity.secret': 'sekrit'}

        self._makeOne(settings)
        self.config.commit()
        introspector = self.config.registry.introspector
        auth_policy = introspector.get('authentication policy', None)

        assert isinstance(auth_policy['policy'], AuthServicePolicy)
        assert verifyClass(IAuthSourceService, find_service_factory(
            self.config, IAuthSourceService))

    def test_include_me_session(self):
        from pyramid_authsanity.policy import AuthServicePolicy
        settings = {'authsanity.source': 'session'}

        self._makeOne(settings)
        self.config.commit()
        introspector = self.config.registry.introspector
        auth_policy = introspector.get('authentication policy', None)

        assert isinstance(auth_policy['policy'], AuthServicePolicy)
        assert verifyClass(IAuthSourceService, find_service_factory(
            self.config, IAuthSourceService))



import pytest

import pyramid.testing
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.interfaces import IAuthenticationPolicy

from zope.interface import implementer

from zope.interface.verify import (
    verifyClass,
    verifyObject
    )

from pyramid_authsanity.interfaces import (
    IAuthSourceService,
    IAuthService,
    )

def test_clean_principal_invalid():
    from pyramid_authsanity.policy import _clean_principal
    from pyramid.security import Everyone

    ret = _clean_principal(Everyone)

    assert ret is None

def test_clean_principal_valid():
    from pyramid_authsanity.policy import _clean_principal

    ret = _clean_principal('root')

    assert ret == 'root'

class TestAuthServicePolicyInterface(object):
    def test_verify(self):
        from pyramid_authsanity.policy import AuthServicePolicy
        assert verifyClass(IAuthenticationPolicy, AuthServicePolicy)
        assert verifyObject(IAuthenticationPolicy, AuthServicePolicy())

class TestAuthServicePolicy(object):
    @pytest.fixture(autouse=True)
    def pyramid_config(self, request):
        from pyramid.interfaces import IDebugLogger
        self.config = pyramid.testing.setUp()
        self.config.set_authorization_policy(ACLAuthorizationPolicy())
        self.logger = DummyLogger()
        self.config.registry.registerUtility(self.logger, IDebugLogger)

        def finish():
            del self.config
            pyramid.testing.tearDown()
        request.addfinalizer(finish)

    def _makeOne(self, debug=False, source=None, auth=None):
        from pyramid_authsanity import AuthServicePolicy
        def find_services(request):
            return (source, auth)

        def session_registered(request):
            return False

        policy = AuthServicePolicy(debug=debug)
        policy._find_services = find_services
        policy._session_registered = session_registered
        return policy

    def _makeOneRequest(self):
        request = DummyRequest()
        request.registry = self.config.registry
        return request

    def test_find_services(self):
        policy = self._makeOne()
        request = self._makeOneRequest()

        (source, auth) = policy._find_services(request)

        assert source is None
        assert auth is None

    def test_fake_source_ticket(self):
        context = None
        request = self._makeOneRequest()
        source = fake_source_init(None)(context, request)
        auth = fake_auth_init()(context, request)

        assert verifyObject(IAuthService, auth)
        assert verifyObject(IAuthSourceService, source)

    def test_valid_source_ticket(self):
        context = None
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid'])(context, request)
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])(context, request)

        policy = self._makeOne(debug=True, source=source, auth=auth)

        authuserid = policy.authenticated_userid(request)
        assert authuserid == 'test'

    def test_invalid_source_ticket(self):
        context = None
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'invalid'])(context, request)
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])(context, request)

        policy = self._makeOne(source=source, auth=auth)

        authuserid = policy.authenticated_userid(request)
        assert authuserid is None

    def test_bad_auth_service(self):
        context = None
        request = self._makeOneRequest()
        source = fake_source_init([None, None])(context, request)

        class BadAuth(object):
            def userid(self):
                raise ValueError('No ticket verified')

            def verify_ticket(self, principal, ticket):
                pass

        auth = BadAuth()

        policy = self._makeOne(source=source, auth=auth)

        authuserid = policy.authenticated_userid(request)
        assert authuserid is None

    def test_no_user_effective_principals(self):
        from pyramid.security import Everyone
        context = None
        request = self._makeOneRequest()
        source = fake_source_init([None, None])(context, request)
        auth = fake_auth_init()(context, request)

        policy = self._makeOne(source=source, auth=auth)

        groups = policy.effective_principals(request)

        assert [Everyone] == groups

    def test_user_bad_principal_effective_principals(self):
        from pyramid.security import Everyone
        context = None
        request = self._makeOneRequest()
        source = fake_source_init([Everyone, 'valid'])(context, request)
        auth = fake_auth_init(fake_userid=Everyone, valid_tickets=['valid'])(context, request)

        policy = self._makeOne(source=source, auth=auth)

        groups = policy.effective_principals(request)

        assert [Everyone] == groups

    def test_effective_principals(self):
        from pyramid.security import Everyone, Authenticated
        context = None
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid'])(context, request)
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])(context, request)

        policy = self._makeOne(source=source, auth=auth)

        groups = policy.effective_principals(request)

        assert [Everyone, Authenticated, 'test'] == groups

    def test_remember(self):
        context = None
        request = self._makeOneRequest()
        source = fake_source_init([None, None])(context, request)
        auth = fake_auth_init(fake_userid='test')(context, request)

        policy = self._makeOne(source=source, auth=auth)

        headers = policy.remember(request, 'test')

        assert len(headers) == 0
        assert len(auth.valid_tickets) >= 1
        assert isinstance(source.value, list)
        assert len(source.value) == 2

    def test_forget(self):
        context = None
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid'])(context, request)
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])(context, request)

        policy = self._makeOne(source=source, auth=auth)

        assert 'valid' in auth.valid_tickets

        headers = policy.forget(request)

        assert len(headers) == 0
        assert 'valid' not in auth.valid_tickets


class TestAuthServicePolicyIntegration(object):
    @pytest.fixture(autouse=True)
    def pyramid_config(self, request):
        from pyramid.interfaces import IDebugLogger, ISessionFactory
        self.config = pyramid.testing.setUp()
        self.config.registry.registerUtility(lambda: None, ISessionFactory)
        self.config.include('pyramid_services')
        self.config.set_authorization_policy(ACLAuthorizationPolicy())
        self.logger = DummyLogger()
        self.config.registry.registerUtility(self.logger, IDebugLogger)

        def finish():
            del self.config
            pyramid.testing.tearDown()
        request.addfinalizer(finish)

    def _makeOne(self, debug=False, source=None, auth=None):
        from pyramid_authsanity import AuthServicePolicy

        if source:
            self.config.register_service_factory(source, iface=IAuthSourceService)

        if auth:
            self.config.register_service_factory(auth, iface=IAuthService)

        return AuthServicePolicy(debug=debug)

    def _makeOneRequest(self):
        from pyramid_services import find_service
        from zope.interface.adapter import AdapterRegistry
        import types

        request = DummyRequest()
        request.find_service = types.MethodType(find_service, request)
        request.registry = self.config.registry
        request.service_cache = AdapterRegistry()

        return request

    def test_include_me(self):
        from pyramid_authsanity.policy import AuthServicePolicy
        self.config.include('pyramid_authsanity')
        self.config.commit()
        introspector = self.config.registry.introspector
        auth_policy = introspector.get('authentication policy', None)

        assert isinstance(auth_policy['policy'], AuthServicePolicy)

    def test_logging(self):
        policy = self._makeOne(debug=True)
        request = DummyRequest(registry=self.config.registry)
        policy._log('this message', 'test_logging', request)

        assert len(self.logger.logentries) >= 1

    def test_find_services(self):
        from pyramid_authsanity.interfaces import (IAuthSourceService, IAuthService)
        self.config.register_service_factory(lambda x, y: 'Source', iface=IAuthSourceService)
        self.config.register_service_factory(lambda x, y: 'Auth', iface=IAuthService)

        policy = self._makeOne()
        request = self._makeOneRequest()

        (sourcesvc, authsvc) = policy._find_services(request)

        assert sourcesvc == 'Source'
        assert authsvc == 'Auth'

    def test_valid_source_ticket(self):
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid'])
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])

        policy = self._makeOne(debug=True, source=source, auth=auth)

        authuserid = policy.authenticated_userid(request)
        assert authuserid == 'test'

    def test_invalid_source_ticket(self):
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'invalid'])
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])

        policy = self._makeOne(source=source, auth=auth)

        authuserid = policy.authenticated_userid(request)
        assert authuserid is None

    def test_bad_auth_service(self):
        request = self._makeOneRequest()
        source = fake_source_init([None, None])

        class BadAuth(object):
            def __init__(self, context, request):
                pass

            def userid(self):
                raise ValueError('No ticket verification')

            def verify_ticket(self, principal, ticket):
                pass

        policy = self._makeOne(source=source, auth=BadAuth)

        authuserid = policy.authenticated_userid(request)
        assert authuserid is None

    def test_no_user_effective_principals(self):
        from pyramid.security import Everyone
        request = self._makeOneRequest()
        source = fake_source_init([None, None])
        auth = fake_auth_init()

        policy = self._makeOne(source=source, auth=auth)

        groups = policy.effective_principals(request)

        assert [Everyone] == groups

    def test_user_bad_principal_effective_principals(self):
        from pyramid.security import Everyone
        request = self._makeOneRequest()
        source = fake_source_init([Everyone, 'valid'])
        auth = fake_auth_init(fake_userid=Everyone, valid_tickets=['valid'])

        policy = self._makeOne(source=source, auth=auth)

        groups = policy.effective_principals(request)

        assert [Everyone] == groups

    def test_effective_principals(self):
        from pyramid.security import Everyone, Authenticated
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid'])
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])

        policy = self._makeOne(source=source, auth=auth)

        groups = policy.effective_principals(request)

        assert [Everyone, Authenticated, 'test'] == groups

    def test_remember(self):
        request = self._makeOneRequest()
        source = fake_source_init([None, None])
        auth = fake_auth_init(fake_userid='test')

        policy = self._makeOne(source=source, auth=auth)

        headers = policy.remember(request, 'test')

        authreq = request.find_service(IAuthService)

        assert len(headers) == 0
        assert len(authreq.valid_tickets) >= 1

    def test_remember_value_json_serializable(self):
        request = self._makeOneRequest()
        source = fake_source_init([None, None])
        auth = fake_auth_init(fake_userid='test')

        policy = self._makeOne(source=source, auth=auth)

        headers = policy.remember(request, 'test')

        authreq = request.find_service(IAuthService)
        sourcereq = request.find_service(IAuthSourceService)

        assert len(headers) == 0
        assert len(authreq.valid_tickets) >= 1
        import json

        assert json.dumps(sourcereq.value)

    def test_remember_same_user(self):
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid_ticket'])
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid_ticket'])

        policy = self._makeOne(source=source, auth=auth)

        headers = policy.remember(request, 'test')

        authreq = request.find_service(IAuthService)

        assert len(headers) == 0
        assert len(authreq.valid_tickets) >= 1

    def test_forget(self):
        request = self._makeOneRequest()
        source = fake_source_init(['test', 'valid'])
        auth = fake_auth_init(fake_userid='test', valid_tickets=['valid'])

        policy = self._makeOne(source=source, auth=auth)

        authreq = request.find_service(IAuthService)

        assert 'valid' in authreq.valid_tickets

        headers = policy.forget(request)

        assert len(headers) == 0
        assert 'valid' not in authreq.valid_tickets

class DummyLogger(object):
    def debug(self, log):
        self.logentries.append(log)

    def __init__(self):
        self.logentries = []

class DummyRequest(object):
    domain = 'example.net'

    def __init__(self, environ=None, session=None, registry=None, cookie=None):
        class Session(dict):
            def invalidate(self):
                self.clear()
            
            def new_csrf_token(self):
                pass

        self.environ = environ or {}
        self.session = Session()
        self.session.update(session or {})
        self.registry = registry
        self.callbacks = []
        self.cookies = cookie or []
        self.context = None


    def add_response_callback(self, callback):
        self.callbacks.append(callback)

def fake_source_init(fake_value):
    @implementer(IAuthSourceService)
    class fake_source(object):
        vary = ['Cookie']
        def __init__(self, context, request):
            self.value = fake_value

        def get_value(self):
            return self.value if self.value else [None, None]

        def headers_remember(self, value):
            self.value = value
            return []

        def headers_forget(self):
            self.value = [None, None]
            return []

    return fake_source


def fake_auth_init(fake_userid=None, fake_groups=list(), valid_tickets=list()):
    @implementer(IAuthService)
    class fake_auth(object):

        def __init__(self, context, request):
            self.authcomplete = False
            self.ticketvalid = False
            self._userid = fake_userid
            self._groups = fake_groups
            self.valid_tickets = valid_tickets

        def userid(self):
            if not self.authcomplete:
                raise ValueError('No ticket verified.')

            if not self.ticketvalid:
                return None

            return self._userid

        def groups(self):
            return self._groups

        def verify_ticket(self, principal, ticket):
            self.authcomplete = True
            if principal == self._userid and ticket in self.valid_tickets:
                self.ticketvalid = True

        def add_ticket(self, principal, ticket):
            if ticket not in self.valid_tickets:
                self.valid_tickets.append(ticket)

        def remove_ticket(self, ticket):
            if ticket in self.valid_tickets:
                self.valid_tickets.remove(ticket)

    return fake_auth



from pyramid.settings import (
    asbool,
    aslist,
    )

from .interfaces import (
    IAuthSourceService,
    )

from .policy import (
    AuthServicePolicy,
    )

from .sources import (
    SessionAuthSourceInitializer,
    CookieAuthSourceInitializer,
    )

from .util import (
    int_or_none,
    kw_from_settings,
    )

default_settings = (
    ('source', str, ''),
    ('cookie.cookie_name', str, 'auth'),
    ('cookie.max_age', int_or_none, None),
    ('cookie.httponly', asbool, True),
    ('cookie.path', str, '/'),
    ('cookie.domains', aslist, []),
    ('cookie.debug', asbool, False),
    ('session.value_key', str, 'sanity.'),
)

def init_cookie_source(config, settings):
    if 'authsanity.secret' not in settings:
        raise RuntimeError('authsanity.secret is required for cookie based storage')

    kw = kw_from_settings(settings, 'authsanity.cookie.')

    config.register_service_factory(CookieAuthSourceInitializer(settings['authsanity.secret'], **kw), iface=IAuthSourceService)

def init_session_source(config, settings):
    kw = kw_from_settings(settings, 'authsanity.session.')

    config.register_service_factory(SessionAuthSourceInitializer(**kw), iface=IAuthSourceService)


default_sources = {
    'cookie': init_cookie_source,
    'session': init_session_source,
}

# Stolen from pyramid_debugtoolbar
def parse_settings(settings):
    parsed = {}

    def populate(name, convert, default):
        name = '%s%s' % ('authsanity.', name)
        value = convert(settings.get(name, default))
        parsed[name] = value
    for name, convert, default in default_settings:
        populate(name, convert, default)
    return parsed


def includeme(config):
    # Go parse the settings
    settings = parse_settings(config.registry.settings)

    # Update the config
    config.registry.settings.update(settings)

    # include pyramid_services
    config.include('pyramid_services')

    if settings['authsanity.source'] in default_sources:
        default_sources[settings['authsanity.source']](config, config.registry.settings)

    config.set_authentication_policy(AuthServicePolicy())



import sys

import pytest

from collections import Iterable

from pyramid_authsanity.interfaces import IAuthSourceService
from pyramid_authsanity import sources

from zope.interface.verify import verifyObject

class _TestAuthSource(object):
    def test_verify_object(self):
        assert verifyObject(IAuthSourceService, self._makeOne())

    def test_get_value(self):
        source = self._makeOne()
        val = source.get_value()

        assert val == [None, None]

    def test_headers_remember(self):
        source = self._makeOne()
        headers = source.headers_remember("test")

        assert isinstance(headers, Iterable)

    def test_headers_forget(self):
        source = self._makeOne()
        headers = source.headers_forget()

        assert isinstance(headers, Iterable)


class TestSessionAuthSource(_TestAuthSource):
    def _makeOne(self, request=None):
        obj = sources.SessionAuthSourceInitializer()

        if request is None:
            request = DummyRequest()

        return obj(None, request)

    def test_remember(self):
        request = DummyRequest()
        source = self._makeOne(request=request)
        source.headers_remember("test")

        assert request.session['sanity.value'] == "test"

    def test_forget(self):
        request = DummyRequest()
        request.session['sanity.value'] = "test"
        source = self._makeOne(request=request)
        headers = source.headers_forget()

        assert 'sanity.value' not in request.session

    def test_get_value_from_session(self):
        request = DummyRequest()
        request.session['sanity.value'] = "test"
        source = self._makeOne(request=request)
        val = source.get_value()

        assert val == "test"

    def test_get_previous_value_after_remember(self):
        request = DummyRequest()
        source = self._makeOne(request=request)
        val = source.get_value()

        assert val == [None, None]

        source.headers_remember("test")

        val2 = source.get_value()

        assert val2 == [None, None]
        assert request.session['sanity.value'] == "test"

class TestCookieAuthSource(_TestAuthSource):
    def _makeOne(self, request=None):
        obj = sources.CookieAuthSourceInitializer('seekrit')

        if request is None:
            request = DummyRequest()

        return obj(None, request)

    def test_get_header_remember(self):
        source = self._makeOne()
        headers = source.headers_remember("test")

        assert isinstance(headers, Iterable)

        for h in headers:
            assert 'auth' in h[1]

    @pytest.mark.skipif(sys.version_info < (3,0),
                    reason="json.dumps() doesn't like binary data on Python 3.x")
    def test_get_header_remember_binary(self):
        source = self._makeOne()
        with pytest.raises(TypeError):
            headers = source.headers_remember(b"test")

    def test_get_header_forget(self):
        source = self._makeOne()
        headers = source.headers_forget()

        assert isinstance(headers, Iterable)

        # Should set an empty cookie
        for h in headers:
            assert h[1].startswith("auth=;")

    def test_get_value_cookie(self):
        request = DummyRequest()
        request.cookies['auth'] = "JgEICiZyfFFc3Qcx5O84h4u8NSZIi51xVMYs_HyP94BO1aXGZpME_LJ1UZgfdAMJDoaGaLCt_y-x6FSBh3ZKDyJ0ZXN0Ig"
        source = self._makeOne(request=request)
        val = source.get_value()

        assert val == "test"

    def test_get_value_bad_cookie(self):
        request = DummyRequest()
        request.cookies['auth'] = "jxxxxxxxfFFc3Qcx5O84h4u8NSZIi51xVMYs_HyP94BO1aXGZpME_LJ1UZgfdAMJDoaGaLCt_y-x6FSBh3ZKDyJ0ZXN0Ig"
        source = self._makeOne(request=request)
        val = source.get_value()

        assert val == [None, None]

    def test_get_value_empty_cookie(self):
        request = DummyRequest()
        request.cookies['auth'] = ""
        source = self._makeOne(request=request)
        val = source.get_value()

        assert val == [None, None]

    def test_round_trip_cookie(self):
        source1 = self._makeOne()
        headers1 = source1.headers_remember(['user1', 'ticket1'])

        assert isinstance(headers1, Iterable)
        assert len(headers1) == 1

        set_cookie = headers1[0][1]
        authpart = set_cookie.split(' ')[0]
        (name, cookie) = authpart.split('=')

        assert name == "auth"

        request = DummyRequest()
        request.cookies[name] = cookie[:-1]

        source2 = self._makeOne(request=request)
        val = source2.get_value()

        assert val == ['user1', 'ticket1']


class DummyRequest(object):
    def __init__(self):
        self.session = dict()
        self.domain = 'example.net'
        self.cookies = dict()




from pytest import raises

from pyramid_authsanity.util import add_vary_callback

class TestAddVaryCallback(object):
    def _makeOne(self, *varies):
        return add_vary_callback(varies)

    def test_add_single_vary(self):
        cb = self._makeOne('cookie')
        response = DummyResponse()
        cb(None, response)

        assert len(response.vary) == 1
        assert 'cookie' in response.vary

    def test_add_multiple_vary(self):
        cb = self._makeOne('cookie', 'authorization')
        response = DummyResponse()
        cb(None, response)

        assert len(response.vary) == 2
        assert 'cookie' in response.vary
        assert 'authorization' in response.vary

    def test_add_multiple_existing(self):
        cb = self._makeOne('cookie')
        response = DummyResponse()
        response.vary = ['cookie']
        cb(None, response)

        assert len(response.vary) == 1
        assert 'cookie' in response.vary

def test_int_or_none_none():
    from pyramid_authsanity.util import int_or_none

    assert None == int_or_none(None)

def test_int_or_none_int():
    from pyramid_authsanity.util import int_or_none

    assert 1 == int_or_none(1)

def test_int_or_none_fail():
    from pyramid_authsanity.util import int_or_none

    with raises(ValueError):
        int_or_none('test')

def test_kw_from_settings():
    from pyramid_authsanity.util import kw_from_settings

    settings = {
            'authsanity.test': None,
            'authsanity.other': 'other',
            'notsanity.test': True,
            }

    kw = kw_from_settings(settings)

    assert kw == {'test': None, 'other': 'other'}

def test_kw_from_settings_custom():
    from pyramid_authsanity.util import kw_from_settings

    settings = {
                'authsanity.cookie.test': True,
                'authsanity.session.test': False,
            }
    kw = kw_from_settings(settings, 'authsanity.cookie.')
    assert kw == { 'test': True }


class DummyResponse(object):
    vary = None



from webob.cookies import (
    SignedCookieProfile,
    )

from zope.interface import implementer

from .interfaces import (
    IAuthSourceService,
    )

def SessionAuthSourceInitializer(value_key='sanity.'):
    """ An authentication source that uses the current session """

    value_key = value_key + 'value'

    @implementer(IAuthSourceService)
    class SessionAuthSource(object):
        vary = []

        def __init__(self, context, request):
            self.request = request
            self.session = request.session
            self.cur_val = None

        def get_value(self):
            if self.cur_val is None:
                self.cur_val = self.session.get(value_key, [None, None])

            return self.cur_val

        def headers_remember(self, value):
            if self.cur_val is None:
                self.cur_val = self.session.get(value_key, [None, None])

            self.session[value_key] = value
            return []

        def headers_forget(self):
            if self.cur_val is None:
                self.cur_val = self.session.get(value_key, [None, None])

            if value_key in self.session:
                del self.session[value_key]
            return []

    return SessionAuthSource


def CookieAuthSourceInitializer(
    secret,
    cookie_name='auth',
    secure=False,
    max_age=None,
    httponly=False,
    path="/",
    domains=None,
    debug=False,
    hashalg='sha512',
    ):
    """ An authentication source that uses a unique cookie. """

    @implementer(IAuthSourceService)
    class CookieAuthSource(object):
        vary = ['Cookie']

        def __init__(self, context, request):
            self.domains = domains

            if self.domains is None:
                self.domains = []
                self.domains.append(request.domain)

            self.cookie = SignedCookieProfile(
                secret,
                'authsanity',
                cookie_name,
                secure=secure,
                max_age=max_age,
                httponly=httponly,
                path=path,
                domains=domains,
                hashalg=hashalg,
                )
            # Bind the cookie to the current request
            self.cookie = self.cookie.bind(request)

        def get_value(self):
            val = self.cookie.get_value()

            if val is None:
                return [None, None]

            return val

        def headers_remember(self, value):
            return self.cookie.get_headers(value, domains=self.domains)

        def headers_forget(self):
            return self.cookie.get_headers(None, max_age=0)

    return CookieAuthSource



from pyramid.interfaces import (
    ISessionFactory,
    )

from .interfaces import (
    IAuthService,
    IAuthSourceService,
    )

def int_or_none(x):
    return int(x) if x is not None else x

def kw_from_settings(settings, from_prefix='authsanity.'):
    return { k.replace(from_prefix, ''): v for k, v in settings.items() if k.startswith(from_prefix) }

def add_vary_callback(vary_by):
    def vary_add(request, response):
        vary = set(response.vary if response.vary is not None else [])
        vary |= set(vary_by)
        response.vary = list(vary)
    return vary_add

def _find_services(request):
    sourcesvc = request.find_service(IAuthSourceService)
    authsvc = request.find_service(IAuthService)

    return (sourcesvc, authsvc)

def _session_registered(request):
    registry = request.registry
    factory = registry.queryUtility(ISessionFactory)

    return (False if factory is None else True)



import base64
import os

from pyramid.interfaces import (
    IAuthenticationPolicy,
    IDebugLogger,
    )

from pyramid.security import (
    Authenticated,
    Everyone,
    )

from .util import (
    add_vary_callback,
    _find_services,
    _session_registered,
    )

from zope.interface import implementer

def _clean_principal(princid):
    """ Utility function that cleans up the passed in principal
    This can easily also be extended for example to make sure that certain
    usernames are automatically off-limits.
    """
    if princid in (Authenticated, Everyone):
        princid = None
    return princid


_marker = object()


@implementer(IAuthenticationPolicy)
class AuthServicePolicy(object):
    def _log(self, msg, methodname, request):
        logger = request.registry.queryUtility(IDebugLogger)
        if logger:
            cls = self.__class__
            classname = cls.__module__ + '.' + cls.__name__
            methodname = classname + '.' + methodname
            logger.debug(methodname + ': ' + msg)

    _find_services = staticmethod(_find_services) # Testing
    _session_registered = staticmethod(_session_registered) # Testing
    _have_session = _marker

    def __init__(self, debug=False):
        self.debug = debug

    def unauthenticated_userid(self, request):
        """ We do not allow the unauthenticated userid to be used. """

    def authenticated_userid(self, request):
        """ Returns the authenticated userid for this request. """
        debug = self.debug

        (sourcesvc, authsvc) = self._find_services(request)
        request.add_response_callback(add_vary_callback(sourcesvc.vary))

        try:
            userid = authsvc.userid()
        except:
            debug and self._log('authentication has not yet been completed',
                    'authenticated_userid', request)
            (principal, ticket) = sourcesvc.get_value()

            debug and self._log(
                'source service provided information: (principal: %r, ticket: %r)'
                % (principal, ticket), 'authenticated_userid', request)

            # Verify the principal and the ticket, even if None
            authsvc.verify_ticket(principal, ticket)

            try:
                # This should now return None or the userid
                userid = authsvc.userid()
            except:
                userid = None

        debug and self._log('authenticated_userid returning: %r' % (userid,),
                'authenticated_userid', request)

        return userid

    def effective_principals(self, request):
        """ A list of effective principals derived from request. """
        debug = self.debug
        effective_principals = [Everyone]

        userid = self.authenticated_userid(request)
        (_, authsvc) = self._find_services(request)

        if userid is None:
            debug and self._log(
                'authenticated_userid returned %r; returning %r' % (
                    userid, effective_principals),
                'effective_principals',
                request
                )
            return effective_principals

        if _clean_principal(userid) is None:
            debug and self._log(
                ('authenticated_userid returned disallowed %r; returning %r '
                 'as if it was None' % (userid, effective_principals)),
                'effective_principals',
                request
                )
            return effective_principals

        effective_principals.append(Authenticated)
        effective_principals.append(userid)
        effective_principals.extend(authsvc.groups())

        debug and self._log(
            'returning effective principals: %r' % (
                effective_principals,),
            'effective_principals',
            request
        )
        return effective_principals

    def remember(self, request, principal, **kw):
        """ Returns a list of headers that are to be set from the source service. """
        debug = self.debug

        if self._have_session is _marker:
            self._have_session = self._session_registered(request)

        prev_userid = self.authenticated_userid(request)

        (sourcesvc, authsvc) = self._find_services(request)

        request.add_response_callback(add_vary_callback(sourcesvc.vary))

        value = {}
        value['principal'] = principal
        value['ticket'] = ticket = base64.urlsafe_b64encode(os.urandom(32)).rstrip(b"=").decode('ascii')

        debug and self._log('Remember principal: %r, ticket: %r' % (principal, ticket), 'remember', request)

        authsvc.add_ticket(principal, ticket)

        # Clear the previous session
        if self._have_session:
            if prev_userid != principal:
                request.session.invalidate()
            else:
                # We are logging in the same user that is already logged in, we
                # still want to generate a new session, but we can keep the
                # existing data
                data = dict(request.session.items())
                request.session.invalidate()
                request.session.update(data)
                request.session.new_csrf_token()

        return sourcesvc.headers_remember([principal, ticket])

    def forget(self, request):
        """ A list of headers which will delete appropriate cookies."""
        debug = self.debug

        if self._have_session is _marker:
            self._have_session = self._session_registered(request)

        (sourcesvc, authsvc) = self._find_services(request)

        request.add_response_callback(add_vary_callback(sourcesvc.vary))

        (_, ticket) = sourcesvc.get_value()

        debug and self._log('Forgetting ticket: %r' % (ticket,), 'forget', request)
        authsvc.remove_ticket(ticket)

        # Clear the session by invalidating it
        if self._have_session:
            request.session.invalidate()

        return sourcesvc.headers_forget()



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()
except IOError:
    README = CHANGES = ''

requires = [
    'pyramid',
    'zope.interface',
    'pyramid_services>=0.3'
]

tests_require = requires + [
    'pytest',
]

coverage_require = tests_require + [
    'coverage',
    'pytest-cov',
    ]

docs_require = requires + [
    'sphinx',
    'repoze.sphinx.autointerface',
    ]

setup(
    name='pyramid_authsanity',
    version='1.0.0',
    description='An auth policy for the Pyramid Web Framework with sane defaults.',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Pyramid',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    keywords='pyramid authorization policy',
    author='Bert JW Regeer',
    author_email='bertjw@regeer.org',
    url='https://github.com/usingnamespace/pyramid_authsanity',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    extras_require={
        'testing': tests_require,
        'coverage': coverage_require,
        'docs': docs_require,
    },
    tests_require=tests_require,
    test_suite='pyramid_authsanity',
    entry_points={
    },
)



from zope.interface import (
    Attribute,
    Interface,
    )

class IAuthSourceService(Interface):
    """ Represents an authentication source. """

    vary = Attribute("List of HTTP headers to Vary the response by.")

    def get_value():
        """ Returns the opaque value that was stored. """

    def headers_remember(value):
        """ Returns any and all headers for remembering the value, as a list.
        Value is a standard Python type that shall be serializable using
        JSON. """

    def headers_forget():
        """ Returns any and all headers for forgetting the current requests
        value. """

class IAuthService(Interface):
    """ Represents an authentication service. This service verifies that the
    users authentication ticket is valid and returns groups the user is a
    member of. """

    def userid():
        """ Return the current user id, None, or raise an error. Raising an
        error is used when no attempt to verify a ticket has been made yet and
        signifies that the authentication policy should attempt to call
        ``verify_ticket``"""

    def groups():
        """ Returns the groups for the current user, as a list. Including the
        current userid in this list is not required, as it will be implicitly
        added by the authentication policy. """

    def verify_ticket(principal, ticket):
        """ Verify that the principal matches the ticket given. """

    def add_ticket(principal, ticket):
        """ Add a new ticket for the principal. If there is a failure, due to a
        missing/non-existent principal, or failure to add ticket for principal,
        should raise an error """

    def remove_ticket(ticket):
        """ Remove a ticket for the current user. Upon success return True """
